#!/usr/bin/env python

import sys

opcodeSet = {
"WRT": 0,
"NOT": 1,
"WRZ": 2,
"WNZ": 3,
"OR": 4,
"AND": 5,
"XOR": 6,
"BNT": 7,
"GRE": 8,
"GRES": 9,
"EQU": 10,
"ADD": 11,
"SUB": 12,
"BSL": 13,
"BSR": 14,
"INC": 15,
"DEC": 16,
"SLO": 17,
"SRO": 18,
"FNC": 19,
"RET": 20,
"FIL": 21,
"FND": 22,
"CII": 23,
"CIIS": 24,
"MUL": 25,
"MULS": 26,
"DIV": 27,
"DIVS": 28,
"MOD": 29,
"ADDF": 30,
"SUBF": 31,
"MULF": 32,
"DIVF": 33,
"CIF": 34,
"CIFS": 35,
"CFI": 36,
"CFIS": 37,
"CIT": 38,
"CITS": 39,
"CFT": 40,
"CTI": 41,
"CTIS": 42,
"CTF": 43,
"WRTT": 44,
"EQUT": 45,
"FNDT": 46,
"LENT": 47,
"SUBT": 48,
"CONT": 49,
"SDL": 50,
"SDS": 51,
"RAN": 52,
"SIN": 53,
"TWR": 54,
"TRD": 55,
"TSL": 56,
"DCL": 57,
"DWR": 58,
"DRD": 59,
"CKT": 60,
"CTK": 61,
"KIP": 62,
"PKY": 63,
"PTX": 64,
"PSL": 65,
"PFL": 66,
"XSM": 67,
"XDR": 68,
"XDW": 69,
"XAR": 70,
"RNF": 71,
"RGAI": 72,
"RGAN": 73,
"RGAA": 74,
"RCR": 75,
"RRD": 76,
"RWR": 77,
"RGN": 78,
"RGS": 79,
"RDL": 80,
"RRN": 81,
"QUIT": 82,
"CTO": 83,
"CIH": 84,
"CHI": 85,
"ALLO": 86,
"FREE": 87,
"DSC": 88,
"DDX": 89,
"DDY": 90,
"GMLF": 91,
"MNDL": 92,
"POW": 93,
"LOG": 94
}

operatorSet = {
"#": 0x00,
"@": 0x10,
"%": 0x20,
"$": 0x30,
"PC": 0x40,
"GS": 0x50,
"LS": 0x60
}

pointerSize = 2

class CodeException(Exception):
    def __init__(self, message, lineNumber):
        self.message = message
        self.lineNumber = lineNumber
    def convertToText(self, code):
        output = self.message
        if self.lineNumber == None:
            output += "\n(No associated line number)"
        else:
            output += "\nError on line " + str(self.lineNumber) + ":"
            tempList = code.split("\n")
            index = self.lineNumber - 2
            if index < 0:
                index = 0
            while index < self.lineNumber + 3 and index < len(tempList):
                output += "\n"
                tempText = tempList[index]
                if index == self.lineNumber:
                    output += ">>> "
                output += tempText
                index += 1
        return output

def appendShortToList(list, number):
    list.append(number & 0x00FF)
    list.append((number & 0xFF00) >> 8)

def setShortInList(list, index, number):
    list[index] = (number & 0x00FF)
    list[index + 1] = (number & 0xFF00) >> 8

def getElementByName(list, name):
    index = 0
    while index < len(list):
        tempElement = list[index]
        if tempElement.name == name:
            return tempElement
        index += 1
    return None

scopeList = []

class Scope:
    def __init__(self, name):
        self.name = name
        self.size = 7
        self.argumentVariableList = []
        self.codeAddress = 0
        self.infoAddress = 0
        self.invocationAddressList = []
        scopeList.append(self)
    def addInfoToBytecode(self, bytecode):
        self.infoAddress = len(bytecode)
        bytecode.append(pointerSize)
        appendShortToList(bytecode, self.codeAddress)
        appendShortToList(bytecode, self.size)
        bytecode.append(len(self.argumentVariableList))
    def populateInvocations(self, bytecode):
        index = 0
        while index < len(self.invocationAddressList):
            tempAddress = self.invocationAddressList[index]
            setShortInList(bytecode, tempAddress, self.infoAddress)
            index += 1

globalScope = Scope("__GLOBAL__")

def getScopeByName(name):
    return getElementByName(scopeList, name)

variableList = []

class Variable:
    def __init__(self, name, scope, offset, size):
        self.name = name
        self.scope = scope
        self.offset = offset
        self.size = size
        variableList.append(self)

def getVariableByNameAndScope(name, scope):
    index = 0
    while index < len(variableList):
        tempVariable = variableList[index]
        if tempVariable.name == name and tempVariable.scope == scope:
            return tempVariable
        index += 1
    index = 0
    while index < len(variableList):
        tempVariable = variableList[index]
        if tempVariable.name == name and tempVariable.scope == globalScope:
            return tempVariable
        index += 1
    return None

labelList = []

class Label:
    def __init__(self, name):
        self.name = name
        self.address = 0
        self.invocationAddressList = []
        labelList.append(self)
    def populateInvocations(self, bytecode):
        index = 0
        while index < len(self.invocationAddressList):
            tempAddress = self.invocationAddressList[index]
            setShortInList(bytecode, tempAddress, self.address)
            index += 1

def getLabelByName(name):
    return getElementByName(labelList, name)

def splitCode(text):
    tempList = text.split("\n")
    output = []
    index = 0
    while index < len(tempList):
        tempText = tempList[index]
        tempList2 = tempText.split(" ")
        output.append(tempList2)
        index += 1
    return output

def extractDeclarationsFromCode(code):
    output = []
    globalOffset = 7
    localOffset = 7
    scope = globalScope
    isInGlobalScope = True
    index = 0
    while index < len(code):
        tempList = code[index]
        tempElement = None
        if len(tempList) > 0:
            tempElement = tempList[0]
        if tempElement == "F":
            localOffset = 7
            isInGlobalScope = False
            scope = Scope(tempList[1])
            tempIndex = 2
            if not tempIndex < len(tempList):
                raise CodeException("Expecting open bracket.", index)
            while tempList[tempIndex] != "[":
                tempSize = pointerSize
                tempName = tempList[tempIndex]
                tempVariable = Variable(tempName, scope, localOffset, tempSize)
                scope.size += tempSize
                scope.argumentVariableList.append(tempVariable)
                localOffset += tempSize
                tempIndex += 1
                if not tempIndex < len(tempList):
                    raise CodeException("Expecting open bracket.", index)
        if tempElement == "]":
            if len(tempList) != 1:
                raise CodeException("Unexpected symbol.", index)
            scope = globalScope
            isInGlobalScope = True
        if tempElement == "L":
            if len(tempList) < 2:
                raise CodeException("Expecting label name.", index)
            if len(tempList) > 2:
                raise CodeException("Unexpected symbol.", index)
            Label(tempList[1])
        if tempElement == "B":
            if len(tempList) < 3:
                raise CodeException("Expecting variable size and name.", index)
            if len(tempList) > 3:
                raise CodeException("Unexpected symbol.", index)
            tempSize = int(tempList[1])
            tempOffset = 0
            if isInGlobalScope:
                tempOffset = globalOffset
                globalOffset += tempSize
            else:
                tempOffset = localOffset
                localOffset += tempSize
            Variable(tempList[2], scope, tempOffset, tempSize)
            scope.size += tempSize
        else:
            output.append(tempList)
        index += 1
    return output

def addValueToBytecode(bytecode, list, index, scope):
    tempElement = list[index]
    index += 1
    if tempElement in operatorSet:
        tempNumber = operatorSet[tempElement]
        if not index < len(list):
            raise CodeException("Expected size.", None)
        tempTerm = list[index]
        try:
            tempSize = int(tempTerm)
        except ValueError:
            raise CodeException("Could not convert size to integer (" + tempTerm + ").", None)
        index += 1
        tempNumber |= tempSize
        if index < len(list):
            tempElement2 = list[index]
            if tempElement2 == ":":
                index += 1
                tempNumber |= 0x08
        bytecode.append(tempNumber)
        return index
    tempScope = getScopeByName(tempElement)
    if tempScope:
        tempScope.invocationAddressList.append(len(bytecode))
        appendShortToList(bytecode, 0)
        return index
    tempVariable = getVariableByNameAndScope(tempElement, scope)
    if tempVariable:
        appendShortToList(bytecode, tempVariable.offset)
        return index
    tempLabel = getLabelByName(tempElement)
    if tempLabel:
        tempLabel.invocationAddressList.append(len(bytecode))
        appendShortToList(bytecode, 0)
        return index
    if tempElement == "D":
        if not index < len(list):
            raise CodeException("Expected size.", None)
        tempTerm = list[index]
        try:
            tempSize = int(tempTerm)
        except ValueError:
            raise CodeException("Could not convert size to integer (" + tempTerm + ").", None)
        index += 1
        if not index < len(list):
            raise CodeException("Expected value.", None)
        tempTerm = list[index]
        try:
            tempValue = int(tempTerm)
        except ValueError:
            raise CodeException("Could not convert value to integer (" + tempTerm + ").", None)
        index += 1
        tempMask = 0xFF
        tempOffset = 0
        tempIndex = 0
        while tempIndex < tempSize:
            tempNumber = (tempValue & tempMask) >> tempOffset
            bytecode.append(tempNumber)
            tempMask <<= 8
            tempOffset += 8
            tempIndex += 1
        return index
    if tempElement == "H":
        if not index < len(list):
            raise CodeException("Expected value.", None)
        tempText = list[index]
        index += 1
        tempIndex = 0
        while tempIndex < len(tempText):
            tempTerm = tempText[tempIndex:(tempIndex + 2)]
            try:
                tempNumber = int(tempTerm, 16)
            except ValueError:
                raise CodeException("Could not convert value to integer (" + tempTerm + ").", None)
            bytecode.append(tempNumber)
            tempIndex += 2
        return index
    if tempElement.find("\"") == 0:
        tempElement = tempElement[1:]
        while True:
            tempIndex = 0
            while tempIndex < len(tempElement):
                tempCharacter = tempElement[tempIndex:(tempIndex + 1)]
                if tempCharacter == "\"":
                    return index
                tempNumber = ord(tempCharacter)
                bytecode.append(tempNumber)
                tempIndex += 1
            bytecode.append(32)
            tempElement = list[index]
            index += 1
    return index

def compileCodeIntoBytecode(code):
    output = []
    output.append(pointerSize)
    appendShortToList(output, globalScope.size)
    scope = globalScope
    index = 0
    while index < len(code):
        tempList = code[index]
        tempElement = tempList[0]
        if tempElement == "F":
            scope = getScopeByName(tempList[1])
            scope.codeAddress = len(output)
        if tempElement == "]":
            scope = globalScope
        if tempElement == "L":
            tempLabel = getLabelByName(tempList[1])
            tempLabel.address = len(output)
        if tempElement == "C":
            tempIndex = 1
            while tempIndex < len(tempList):
                try:
                    tempIndex = addValueToBytecode(output, tempList, tempIndex, scope)
                except CodeException as exception:
                    exception.lineNumber = index
                    raise exception
        if tempElement in opcodeSet:
            output.append(opcodeSet[tempElement])
            tempIndex = 1
            while tempIndex < len(tempList):
                try:
                    tempIndex = addValueToBytecode(output, tempList, tempIndex, scope)
                except CodeException as exception:
                    exception.lineNumber = index
                    raise exception
        index += 1
    index = 0
    while index < len(scopeList):
        tempScope = scopeList[index]
        if tempScope != globalScope:
            tempScope.addInfoToBytecode(output)
        index += 1
    return output

def populateScopeAndLabelInvocations(bytecode):
    index = 0
    while index < len(scopeList):
        tempScope = scopeList[index]
        tempScope.populateInvocations(bytecode)
        index += 1
    index = 0
    while index < len(labelList):
        tempLabel = labelList[index]
        tempLabel.populateInvocations(bytecode)
        index += 1

print "DUO Light Bytecode Assembler"

if len(sys.argv) < 2:
    print "Please provide a file name as an argument."
    sys.exit(0)

inputFileName = sys.argv[1];
tempFile = open(inputFileName, "r")
inputText = tempFile.read()
tempFile.close()

try:
    code = splitCode(inputText)
    code = extractDeclarationsFromCode(code)
    outputList = compileCodeIntoBytecode(code)
    populateScopeAndLabelInvocations(outputList)
except CodeException as exception:
    print exception.convertToText(inputText)
    sys.exit()

index = inputFileName.rfind(".")
outputFileName = inputFileName[:index] + ".dlbc"
tempFile = open(outputFileName, "w")
tempFile.write(bytearray(outputList))
tempFile.close()

print "Finished assembling."
